# C 编程，第 3 部分：常见问题

> 原文：<https://github.com/angrave/SystemProgramming/wiki/C-Programming%2C-Part-3%3A-Common-Gotchas>

C 程序员常犯的错误是什么？

## 记忆错误

## 字符串常量是常量

```c
char array[] = "Hi!"; // array contains a mutable copy 
strcpy(array, "OK");

char *ptr = "Can't change me"; // ptr points to some immutable memory
strcpy(ptr, "Will not work");
```

字符串文字是存储在程序代码段中的字符数组，它是不可变的。两个字符串文字可以在内存中共享相同的空间。一个例子如下：

```c
char * str1 = "Brandon Chong is the best TA";
char * str2 = "Brandon Chong is the best TA";
```

`str1`和`str2`指向的字符串实际上可能位于内存中的相同位置。

但是，字符数组包含已从代码段复制到堆栈或静态内存中的文字值。以下 char 数组不驻留在内存中的相同位置。

```c
char arr1[] = "Brandon Chong didn't write this";
char arr2[] = "Brandon Chong didn't write this";
```

## 缓冲区溢出/下溢

```c
#define N (10)
int i = N, array[N];
for( ; i >= 0; i--) array[i] = i;
```

C 不检查指针是否有效。上面的例子写入`array[10]`，它位于数组边界之外。这可能会导致内存损坏，因为该内存位置可能正在用于其他内容。在实践中，这可能更难以发现，因为溢出/下溢可能发生在库调用中，例如

```c
gets(array); // Let's hope the input is shorter than my array!
```

## 返回指向自动变量的指针

```c
int *f() {
    int result = 42;
    static int imok;
    return &imok; // OK - static variables are not on the stack
    return &result; // Not OK
}
```

自动变量仅在函数的生命周期内绑定到堆栈内存。函数返回后，继续使用内存是错误的。

## 内存分配不足

```c
struct User {
   char name[100];
};
typedef struct User user_t;

user_t *user = (user_t *) malloc(sizeof(user));
```

在上面的例子中，我们需要为 struct 分配足够的字节。相反，我们分配了足够的字节来保存指针。一旦我们开始使用用户指针，我们将破坏内存。正确的代码如下所示。

```c
struct User {
   char name[100];
};
typedef struct User user_t;

user_t * user = (user_t *) malloc(sizeof(user_t));
```

#### 字符串需要`strlen(s)+1`个字节

每个字符串在最后一个字符后必须有一个空字节。要存储字符串`"Hi"`，需要 3 个字节：`[H] [i] [\0]`。

```c
  char *strdup(const char *input) {  /* return a copy of 'input' */
    char *copy;
    copy = malloc(sizeof(char*));     /* nope! this allocates space for a pointer, not a string */
    copy = malloc(strlen(input));     /* Almost...but what about the null terminator? */
    copy = malloc(strlen(input) + 1); /* That's right. */
    strcpy(copy, input);   /* strcpy will provide the null terminator */
    return copy;
}
```

## 使用未初始化的变量

```c
int myfunction() {
  int x;
  int y = x + 2;
...
```

自动变量保存垃圾（无论位模式发生在内存中）。假设它将始终初始化为零是错误的。

## 假设未初始化的内存将被清零

```c
void myfunct() {
   char array[10];
   char *p = malloc(10);
```

自动（临时变量）不会自动初始化为零。使用 malloc 的堆分配不会自动初始化为零。

## 双免费

```c
  char *p = malloc(10);
  free(p);
//  .. later ...
  free(p); 
```

将同一块内存释放两次是错误的。

## 晃来晃去的指针

```c
  char *p = malloc(10);
  strcpy(p, "Hello");
  free(p);
//  .. later ...
  strcpy(p,"World"); 
```

不应使用释放内存的指针。防御性编程实践是在释放内存后立即将指针设置为 null。

最好将 free 释放到以下片段中，该片段在以下情况下自动将释放的变量设置为 null：（vim - ultisnips）

```source-viml
snippet free "free(something)" b
free(${1});
$1 = NULL;
${2}
endsnippet
```

## 逻辑和程序流程错误

## 忘了休息

```c
int flag = 1; // Will print all three lines.
switch(flag) {
  case 1: printf("I'm printed\n");
  case 2: printf("Me too\n");
  case 3: printf("Me three\n");
}
```

不间断的 Case 语句将继续执行下一个 case 语句的代码。正确的代码如下所示。最后一个语句的中断是不必要的，因为在最后一个语句之后不再需要执行。但是，如果添加更多，则可能会导致一些错误。

```c
int flag = 1; // Will print only "I'm printed\n"
switch(flag) {
  case 1: 
    printf("I'm printed\n");
    break;
  case 2: 
    printf("Me too\n");
    break;
  case 3: 
    printf("Me three\n");
    break; //unnecessary
}
```

## 平等与平等

```c
int answer = 3; // Will print out the answer.
if (answer = 42) { printf("I've solved the answer! It's %d", answer);}
```

## 未声明或不正确的原型功能

```c
time_t start = time();
```

系统函数'time'实际上是一个参数（指向一些可以接收 time_t 结构的内存的指针）。编译器没有捕获此错误，因为程序员没有通过包含`time.h`来提供有效的函数原型

## 额外的分号

```c
for(int i = 0; i < 5; i++) ; printf("I'm printed once");
while(x < 10); x++ ; // X is never incremented
```

但是，以下代码完全可以。

```c
for(int i = 0; i < 5; i++){
    printf("%d\n", i);;;;;;;;;;;;;
}
```

拥有这种代码是可以的，因为 C 语言使用分号（;）来分隔语句。如果分号之间没有语句，则无需执行任何操作，编译器将继续执行下一个语句

## 其他陷阱

## 预处理器

什么是预处理器？这是在实际编译程序之前编译器执行**的操作。它是一个复制和粘贴命令。这意味着如果我执行以下操作。**

```c
#define MAX_LENGTH 10
char buffer[MAX_LENGTH]
```

预处理后，它看起来像这样。

```c
char buffer[10]
```

## C 预处理器宏和副作用

```c
#define min(a,b) ((a)<(b) ? (a) : (b))
int x = 4;
if(min(x++, 100)) printf("%d is six", x);
```

宏是简单的文本替换，因此上面的示例扩展为`x++ &lt; 100 ? x++ : 100`（为清楚起见省略了括号）

## C 预处理器宏和优先级

```c
#define min(a,b) a<b ? a : b
int x = 99;
int r = 10 + min(99, 100); // r is 100!
```

宏是简单的文本替换，因此上面的示例扩展为`10 + 99 &lt; 100 ? 99 : 100`

## C 预处理器逻辑问题

```c
#define ARRAY_LENGTH(A) (sizeof((A)) / sizeof((A)[0]))
int static_array[10]; // ARRAY_LENGTH(static_array) = 10
int* dynamic_array = malloc(10); // ARRAY_LENGTH(dynamic_array) = 2 or 1
```

宏有什么问题？好吧，如果我们有一个像第一个数组那样的静态数组，那么它是有效的，因为静态数组的 sizeof 返回数组占用的字节数，并将它除以 sizeof（an_element）将得到条目数。但是如果我们使用指向一块内存的指针，那么获取指针的大小并将其除以第一个条目的大小并不总能给出数组的大小。

## `sizeof`有什么用吗？

```c
int a = 0;
size_t size = sizeof(a++);
printf("size: %lu, a: %d", size, a);
```

代码打印出来的是什么？

```
size: 4, a: 0 
```

因为 sizeof 实际上并未在运行时进行评估。编译器分配所有表达式的类型并丢弃表达式的额外结果。